--TODO: copyright


local wave, waveObject = {
  overwrite = false,
  
  energyRatio = 1.2, --energy peak detection threshold
  trainSize = 108, --train size for convolution, blocks of 1024 (430 = 10s)
  
  minLength = 2048 --minimum length, in samples, to parse audio
  
}, {}
setmetatable(wave, wave)

--[[ Private ]]--

local tr2 = 1.0594630943592952645

local playInstance, stopInstance, isPlaying

local function removeStopped(sources)
	local remove = {}
	for s in pairs(sources) do
		remove[s] = true
	end
	for s in pairs(remove) do
		sources[s] = nil
	end
end

local function lerp(a, b, k) --smooth transitions
  if a == b then
    return a
  else
    if math.abs(a-b) < 0.005 then return b else return a * (1-k) + b * k end
  end
end

--[[ Public - Wave ]]--

function wave:newSource(path, type)
  local _object = {
		_paused   = false,
		path      = path,
		type      = type,
		instances = {},
		looping   = false,
		pitch     = 1,
		volume    = 1,
        playing = false
	}
  
  setmetatable(_object, { __index = waveObject })
  
  if _object.type == "static" then
    _object:Parse()
  end

  return _object
end

------------------ [{[{[ Music Functions ]}]}] ------------------

-----{{{{ play() }}}}-----
function waveObject:Play(pitched)
    local instance = love.audio.newSource(self.data or self.path, self.type or "stream")
    self.instances[instance] = instance
    
    instance:setLooping(self.looping)
	--instance:setPitch(self.pitch + (pitched and (math.random()-.5)*.1 or 0) )
	instance:setVolume(self.volume)
    
    love.audio.stop(self.instances)
    love.audio.play(instance)
    
    if self:isMusic() or self:isParsed() then
        self.instance = instance
        if self:isMusic() then --start listening to beat
            self.duration = self.duration or self.instance:getDuration()
            self.previousFrame = love.timer.getTime() * 1000
            self.lastTime = 0
            self.time = 0
            self.beat = 0
            self.beatTime = 0
        end
    end
  
    return self
end

-----{{{{ stop() }}}}-----

function waveObject:Stop()
	for s in pairs(self.instances) do
		s:stop()
	end
 
	self.instances = {}
  
    return self
end

-----{{{{ loop() }}}}-----

function waveObject:Loop(A)
	if A == 1 then
        self.looping = true
    elseif A == 0 then
        self.looping = false
    end
    
    return self
end

-----{{{{ volume() }}}}-----

function waveObject:Volume(A)
    self.volume = A
    
    return self
end

-----{{{{ isPlaying() }}}}-----
function waveObject:isPlaying()
      self.playing = false
      for s in pairs(self.instances) do
            if s then self.playing = true end --if there's at least one instance
        end
      return self
end

------------------------------------------------------------------------

------------------ [{[{[ Music Setup ]}]}] ------------------

function waveObject:Parse()
  assert(not self.data, "Audio is already parsed.")
  self.data = love.sound.newSoundData(self.path)
  self.length = self.data:getSampleCount()
  self.duration = self.data:getDuration()
  return self
end

------------------------------------------------------------------------

------------------ [{[{[ Beat Functions ]}]}] ------------------

function waveObject:onBeat(f)
  self.onBeat = f
  return self
end

function waveObject:calculateBeat()
  return (self.bpm / 60) * ((self.time + (self.offset or 0)) % (self.duration * 1000)) / 1000
end

------------------ [{[{[ Energy Functions ]}]}] ------------------

function waveObject:calculateEnergy(offset, size, limiter)
  
  local _energy = 0
  for i = offset, offset + size do
    _energy = _energy + (self.data:getSample(i) ^ 2)/ (limiter or size);
  end
  
  return _energy
end

function waveObject:getEnergy()
  if not self:isParsed() then return 0 end
  return self.energy
end

function waveObject:SetIntensity(intensity)
  self.intensity = intensity
  self.energy = 0
  return self
end

------------------------------------------------------------------------


------------------ [{[{[ BPM Functions ]}]}] ------------------

function waveObject:SetBPM(bpm)
  self.bpm = bpm
  self.bps = 60 / bpm
  self.detector = nil
  return self
end

function waveObject:calculateBPM()

  local _d = self.detector
  
  local _size = _d.size --length/1024

  --calculate instant energy

  for i = 0, _size + 42 do
    _d.e1024[i] = self:calculateEnergy(1024 * i, 4096) --4096 makes it smoother
  end

  --calculate average energy for 1 sec

  _d.e44100[0] = 0

  --average of the first 43 first e1024 gives the average e44100

  local sum = 0

  for i = 0, 43 do
    sum = sum + _d.e1024[i];
  end

  _d.e44100[0] = sum / 43;

  --fill others
  
  for i = 1, _size do
    sum = sum - _d.e1024[i - 1] + _d.e1024[i + 42] --because 42 is the only answer
    _d.e44100[i] = sum / 43;
  end

  --calculate ratio

  for i = 21, _size do

    -- -21 centers e1024 on e44100's second

    if (_d.e1024[i] > wave.energyRatio * _d.e44100[i-21]) then
      
      _d.energyPeak[i] = 1
      
    end

  end

  --calculate BPM

  --calculate time between every energy peak

  local T = {}

  local iPrevious = 0

  for i = 1, _size do

    if (_d.energyPeak[i] == 1) and (_d.energyPeak[i-1] == 0) then

      local di = i - iPrevious

      if (di > 5) then --interferences

        table.insert(T, di)

        iPrevious = i

      end
    end
  end

  --let the statistics begin

  local maxOccT, avgOccT = 0, 0

  --count the frequency of each interval

  local occT = {}; --86 = max 2 blocks of 43 gap (=2s)

  for i = 0, 86 do
    occT[i] = 0
  end

  for i = 1, #T do
    if(T[i] <= 86) then
      occT[T[i]] = occT[T[i]] + 1
    end
  end
  

  local occMax = 0

  for i = 1, 86 do

    if occT[i] > occMax then

      occMaxT = i

      occMax = occT[i]

    end
  end

  --average of max + max neighbour for more precision

  local neighbour = occMaxT - 1

  if (occT[occMaxT + 1] > occT[neighbour]) then
    neighbour = occMaxT + 1
  end

  local div = occT[occMaxT] + occT[neighbour]

  if div == 0 then
    avgOccT = 0
  else
    avgOccT = (occMaxT * occT[occMaxT] + neighbour * occT[neighbour]) / div
  end



  --last step

  local bpm = 60 / (avgOccT * (1024/44100))
  
  while bpm < 90 or bpm > 180 do --clamp at a reasonable bpm
    if bpm < 90 then
      bpm = bpm * 2
    else
      bpm = bpm * .5
    end
  end
  
  return math.floor(bpm + .5)
end

function waveObject:detectBPM()
  assert(self:isParsed(), "Parse audio before enabling beat detection.")
  
  local length = self.length
  local size = math.floor(length/1024)
  
  self.detector = {
    
    length = length,
    e1024 = {}, --size
    e44100 = {}, --size
    energyPeak = {}, --size + 21
    
    size = size
  }
  
  for i=0, size + 21 do
    self.detector.energyPeak[i] = 0
  end
  
  self:setBPM( self:calculateBPM() ) --heavy stuff
  
  return self
end

------------------------------------------------------------------------

------------------ [{[{[ Update Functions ]}]}] ------------------

--// Update

function waveObject:update(dt)
    if self:isMusic() then self:updateBeat(dt) end
    if self:isParsed() then self:updateEnergy(dt) end

    return self
end

function waveObject:updateBeat(dt)
  local _instance = self.instances[self.instance]
  if not _instance then return self end
  
  local _offset = love.timer.getTime() * 1000
  
  local _elapsedBeats = 0
  
  self.time = self.time + _offset - self.previousFrame
  
  self.previousFrame = _offset
  local _position = _instance:tell("seconds") * 1000
  if _position < self.lastTime then --music looped
    self.time = _position
    self.lastTime = _position
    _elapsedBeats = _elapsedBeats + 1
    self.beat = 0
  elseif _position ~= self.lastTime then --updates music time, but with easing
    self.time = (self.time + (_position))/2
    self.lastTime = _position
  end
  
  self.beatTime = self:calculateBeat()
  
  local _beat = math.floor(self.beatTime)
  _elapsedBeats = _elapsedBeats + _beat - self.beat
  self.beat = _beat
  
  if self.onBeat then
    for i = 1, _elapsedBeats do
      self.onBeat()
    end
  end
  
  return self
end

function waveObject:updateEnergy(dt)
  local _instance = self.instances[self.instance]
  if not _instance then return self end
  
  local _sample = _instance:tell( "samples" )
  local size = 1024
  if _sample > size then
    
    local _energy = self:calculateEnergy(_sample, size, self.intensity)
    self.energy = lerp(self.energy, _energy, 10*dt)
    
  end
  return self
end

------------------ [{[{[ Checks ]}]}] ------------------

function waveObject:isMusic() return self.bpm and true or false end
function waveObject:isParsed() return (self.data and self.duration > 10) and true or false end

------------------ [{[{[ Aliases ]}]}] ------------------

waveObject.isLooping = waveObject.getLooping

------------------ [{[{[ End ]}]}] ------------------

return wave
