--TODO: allow changes in speed to the Menu.Snake without it bugging out
--TODO: Fix collisions with edges
--TODO: Add a score count when collecting food
--TODO: Add a death screen instead of just going to title screen
--TODO: Add sound
--TODO: add effects i.e screen shake, screen rotation, background movement, anything trippy
--TODO: Ui effects i.e fade screen on death
--TODO: Smoother graphics
--TODO: Neater Code
--TODO: Score counter in corner
--TODO: Allow "food" to have neater graphics i.e pulsing effect
--TODO: Make better code for background to fit screen (Math shit)

local Menu = require("Menu")
local Beat = require "Beat-D"

------------------------ [ Localizing ] ------------------------

local directionQueue
local foodPosition
local canvas
local BackgroundX, BackgroundY, Background
local width, height
local screen_height, screen_width

---- { Game } ----
local Game = {}

---- { Timer } ----
local timer = 0

---- { Debug } ----
local Debug = 0

---- { Screen Variables } ----
screen_width, screen_height = love.window.getDesktopDimensions()
width, height = love.graphics.getWidth(), love.graphics.getHeight()

------------------------ [ Test Zone ] ------------------------
--local Menu = Beat:newSource("Menu.wav")

------------------------ [ This moves the "food" ] ------------------------
local function moveFood()
    local possibleFoodPositions = {}
    
    for foodX = 1, Menu.gridXCount do
        for foodY = 1, Menu.gridYCount do
            local possible = true
            
            for segmentIndex, segment in ipairs(Menu.Snake.Segments) do
                if foodX == segment.x and foodY == segment.y then
                    possible = false
                end
            end
            
            if possible then
                table.insert(possibleFoodPositions, { x = foodX, y = foodY })
            end
        end
    end
    
    foodPosition = possibleFoodPositions[love.math.random(#possibleFoodPositions)]
end


------------------------ [ This loads the game ] ------------------------
function Game:load()
    ---- { Menu Setup } ----
    Game.AllowMenu = 1
    Game.MenuC = 0
    
    ---- { Font handler } ----
    Menu.font = love.graphics.newFont('PixelFJVerdana12pt_0.ttf', 20)
    love.graphics.setFont(Menu.font)
    
    ---- { canvas } ----
    canvas = love.graphics.newCanvas(love.graphics.getHeight(), love.graphics.getWidth())
    
    ---- { starting direction } ----
    directionQueue = { 'right' }
    
    ---- { background segments } ----
    Background = {
        Segments = {
            { x = 1, y = 1 },
            { x = 2, y = 1 },
            { x = 3, y = 1 },
            { x = 4, y = 1 },
            { x = 5, y = 1 },
            { x = 6, y = 1 },
            { x = 7, y = 1 },
            { x = 8, y = 1 },
            { x = 9, y = 1 },
            { x = 10, y = 1 }
        }
    }
    
    ---- { "food" position } ----
    foodPosition = {
        x = love.math.random(1, Menu.gridXCount),
        y = love.math.random(1, Menu.gridYCount)
    }
    
    moveFood()
    
    ---- { screen setup } ----
    BackgroundX = love.graphics.getWidth() / 15
    BackgroundY = love.graphics.getHeight() / 15
    
    ---- { Menu load } ----
    if Menu.Active == 1 then
        Menu:load()
    end
    
    ---- { background setup } ----
    for a = 1, Menu.squareRows do
        for b = 1, Menu.squareColumns do
            table.insert(Background.Segments, 1, { x = a, y = b })
        end
    end
end

------------------------ [ This updates the game ] ------------------------

function Game:update(dt)
    if Menu.Active == 1 then
        Menu:update()
        
        if Menu.FileC == 1 then
            Game.MenuC = 1
        end
    else
    
        ---- { Playlist + Menu handling } ----

        if Music.Playlist[Game.MenuC].playing == false and Game.AllowMenu == 1 and Game.MenuC > Menu.FileC then
            Music.Playlist[Game.MenuC]:Loop(0)
            Music.Playlist[Game.MenuC]:SetIntensity(100)
            Music.Playlist[Game.MenuC]:Parse()
            Music.Playlist[Game.MenuC]:Play()
            
            Game.AllowMenu = 0
        end
        
        if not Music.Playlist[Game.MenuC].playing == true and Game.AllowMenu == 0 then
            Game.AllowMenu = 1
            if Game.MenuC < Menu.FileC then
                Game.MenuC = Game.MenuC + 1
            end
            
            return Game.MenuC, Music.Playlist[Game.MenuC]
        end
        
        print(Game.MenuC, #Menu.Playlist, Music.Playlist[Game.MenuC])
        
        if Game.MenuC > 0 then
            Music.Playlist[Game.MenuC]:update(dt)
        end
        
        
    
        ---------------------------------------

        timer = timer + dt
        if Menu.Snake.Alive == true then
            local timerLimit = 0.15
            if timer >= timerLimit then
                timer = timer - timerLimit
                
                local nextXPosition = Menu.Snake.Segments[1].x
                local nextYPosition = Menu.Snake.Segments[1].y
                
                ---- { cleanup } ----
                if #directionQueue > 1 then
                    table.remove(directionQueue, 1)
                end
    
                if directionQueue[1] == 'right' then
                    nextXPosition = nextXPosition + Menu.Snake.speed
                elseif directionQueue[1] == 'left' then
                    nextXPosition = nextXPosition - Menu.Snake.speed
                elseif directionQueue[1] == 'down' then
                    nextYPosition = nextYPosition + Menu.Snake.speed
                elseif directionQueue[1] == 'up' then
                    nextYPosition = nextYPosition - Menu.Snake.speed
                end
         
                for segmentIndex, segment in ipairs(Menu.Snake.Segments) do
                    if segmentIndex ~= #Menu.Snake.Segments
                            and nextXPosition == segment.x
                            and nextYPosition == segment.y then
                        Menu.Snake.Alive = false
                    end
                end
                
                ---- { collision detection } ----
                
                if nextYPosition > (Menu.squareColumns - (Menu.squareColumns * 14 - love.graphics.getHeight() + 12)) then
                    Menu.Snake.Alive = false
                end
                if nextXPosition > (Menu.squareRows - (Menu.squareRows * 14 - love.graphics.getWidth() + 14)) then
                    Menu.Snake.Alive = false

                end
                if nextYPosition < 1 then
                    Menu.Snake.Alive = false
                end
                if nextXPosition < 1 then
                    Menu.Snake.Alive = false
                end
                
                ---- { body handling } ----
                table.insert(Menu.Snake.Segments, 1, { x = nextXPosition, y = nextYPosition })
                table.remove(Menu.Snake.Segments)
                table.insert(Menu.Snake.Segments, 1, { x = nextXPosition, y = nextYPosition })
                
                if Menu.Snake.Segments[1].x == foodPosition.x and Menu.Snake.Segments[1].y == foodPosition.y then
                    table.insert(Menu.Snake.Segments, 1, { x = nextXPosition, y = nextYPosition })
                    moveFood()
                else
                    table.remove(Menu.Snake.Segments)
                end
            
            end

        elseif timer >= 2 then
            Menu.Active = 1
            
            table.remove(Menu.Snake.Segments)
            
            love.load()
        end
    
    end
end

------------------------ [ This draws the game ] ------------------------

function Game:draw()
    if Menu.Active == 1 then
        Menu:draw()
    else
        
        ---- { snake draw } ----
        local function drawCell(x, y)
            love.graphics.rectangle(
                    'fill',
                    (x - 1) * Menu.cellSize + Menu.Playlist[0]:getEnergy()*5,
                    (y - 1) * Menu.cellSize + Menu.Playlist[0]:getEnergy()*5,
                    Menu.cellSize - Menu.cellSize + Menu.Playlist[0]:getEnergy()*4,
                    Menu.cellSize - Menu.cellSize + Menu.Playlist[0]:getEnergy()*4
            )
        end
        
        ---- { background draw } ----
        for a = 1, Menu.squareRows do
            for b = 1, Menu.squareColumns do
                love.graphics.setColor(0, .28, .28)
                drawCell(a, b)
            end
        end
        
        ---- { snake color } ----
        for segmentIndex, segment in ipairs(Menu.Snake.Segments) do
            if Menu.Snake.Alive then
                love.graphics.setColor(Menu.Snake.R, Menu.Snake.G, Menu.Snake.B)
            else
                love.graphics.setColor(.5, .5, .5)
            end
            drawCell(segment.x, segment.y)
        end
        
        love.graphics.setColor(1, .3, .3)
        drawCell(foodPosition.x, foodPosition.y)
        
        ---- { debug } ----
        if Debug == 1 then
            for directionIndex, direction in ipairs(directionQueue) do
                love.graphics.setColor(1, 1, 1)
                love.graphics.print('directionQueue[' .. directionIndex .. ']: ' .. direction, 15, 15 * directionIndex)
            end
        else
            love.graphics.print("", 15, 15)
        end
    end
end

------------------------ [ This handles key presses ] ------------------------
function Game:keypressed(key)
    if Menu.Active == 1 then
        Menu.keypressed(key)
    else
        
        ---- { snake movement } ----
        if key == 'right'
                and directionQueue[#directionQueue] ~= 'right'
                and directionQueue[#directionQueue] ~= 'left' then
            table.insert(directionQueue, 'right')
        
        elseif key == 'left'
                and directionQueue[#directionQueue] ~= 'left'
                and directionQueue[#directionQueue] ~= 'right' then
            table.insert(directionQueue, 'left')
        
        elseif key == 'up'
                and directionQueue[#directionQueue] ~= 'up'
                and directionQueue[#directionQueue] ~= 'down' then
            table.insert(directionQueue, 'up')
        
        elseif key == 'down'
                and directionQueue[#directionQueue] ~= 'down'
                and directionQueue[#directionQueue] ~= 'up' then
            table.insert(directionQueue, 'down')
        
        ---- { debug key handle } ----
        elseif key == '`' then
            Debug = Debug + 1
            if Debug == 2 then
                Debug = 0
            end
        end
    end
end

return Game, Menu.Snake
















